/*********************************************************************
 *  File:         POPHELLO.C
 *  Author:       Rick Hollinbeck, 1990
 *  Purpose:      A demonstration program showing how to write an
 *                OS/2 pop-up program.
 *  Compilers:    Microsoft C v5.1, 6.0
 *                Zortech C v2.0
 *  Compile with: (Microsoft)
 *                cl /Lp /G2s pophello.c
 *                The /Gs (disable stack probes) flag is critical, since
 *                this program creates stacks for its threads.
 *
 *                (Zortech)
 *                ztc -b pophello.c (-cpp also works for C++ compatibility)
 *
 *  Link with:
 *                link /NOE/NOI pophello,,,os2 slibcep[,pophello.def];
 *
 *                With either compiler, to use from a windowed command
 *                line (i.e., from within the Presentation Manager's screen
 *                group), the .exe must be marked as WINDOWCOMPAT. You can
 *                do this by either providing a .def file to the linker
 *                that contains the line "NAME pophello WINDOWCOMPAT" or
 *                by using the new exehdr.exe file that comes with MSC
 *                6.0.  To convert pophello.exe to WINDOWCOMPAT, use the
 *                command line "exehdr /pmtype:vio pophello.exe".
 *
 *  Load and run: "detach pophello". You can then pop up "Hello world!"
 *                anytime you are lonely with the CTRL-ALT-T hot key.
 *                Remove by running again with: "pophello".
 *
 *  (c) C Gazette, 1990. May be used freely if source is acknowledged.
 *********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* Load the necessary OS/2 header symbols ... */
#define INCL_DOSMONITORS    /* Device monitor definitions */
#define INCL_DOSINFOSEG     /* DOS information segment definitions */
#define INCL_DOSPROCESS     /* Process control definitions */
#define INCL_DOSSEMAPHORES  /* Semaphore definitions */
#define INCL_KBD            /* Keyboard definitions */
#define INCL_VIO            /* Video definitions */
#define INCL_NOPM           /* No Presentation Manager definitions */
#include <os2.h>            /* OS/2 access definitions */

#define USE_THREAD          /* Define to compile for detached execution */

/*#define KBDTRACE          * Enable debugging keystroke trace to stderr */
/* Run with "detach pophello 2>trace.txt */

/* OS/2 thread main function type */
#define THREAD void far cdecl
#define BOOLEAN int         /* True/False values only */

static BOOLEAN unload_flag = FALSE; /* Set when unloading the process */
static BOOLEAN popped_up = FALSE;   /* Flag when we are popped-up */
static USHORT  waitflag;            /* Flag for VioPopUp() */

#ifdef USE_THREAD

/* DosMonReg constants */
#define NOPREFERENCE    0       /* don't care where in the chain */
#define FRONT           1       /* where in the monitor chain */
#define BACK            2       /* do we want to be installed? */
#define RELEASE         0x40    /* bit mask to distinguish */
/* key press from key release */
#define MINBUFFSIZE 512         /* size for monitor buffers */
/* 64 is minimum, 128 or more is recommended */

/* KBD$ monitor data record */
typedef struct KeyPacket {
    unsigned mnflags;
    KBDKEYINFO cp;
    unsigned ddflags;
} KEYPACKET;

/* Generic monitor buffer header */
typedef struct MonBuffer {
    unsigned bufflen;
    unsigned ddbufflen;
    unsigned char dispatch[12];
} MONBUFFER;

/* Semaphores ... */
/* RAM Semaphore signalling pop-up request */
static HSEM hsemPopup;
/* Signals that we're loaded for all screen groups */
static HSYSSEM ssemLoaded;
/* System Semaphore name prefix */
static char ssem_name[] = "\\sem\\pop";
/* System semaphore name for current screen group */
static char ssemLoaded_name[11];

static USHORT screen_group;             /* Screen group being monitored */
static HMONITOR hmonHandle;             /* Monitor handle */
static MONBUFFER *InBuff, *OutBuff;     /* Buffers for monitor read/writes */
static SEL selGlobalSeg, selLocalSeg;   /* Global and Local seg. selectors */
static GINFOSEG FAR *pgisInfo;          /* Pointer to the Global Info seg.t */
static USHORT keycount;                 /* Size for DosMonRead/DosMonWrite */
static BOOLEAN nested_hotkey = FALSE;   /* True if hotkey while popped-up */
static BOOLEAN in_main_thread = FALSE;  /* True if main thread is active */

/* Keyboard monitor thread */
#define MON_THREAD_STACKSIZE 512    /* Keybd monitor thread's stack size */
static TID mon_thread_id;           /* Keybd monitor thread id */
static BYTE mon_thread_stack[MON_THREAD_STACKSIZE]; /* Space for stack */

/* Exit semaphore monitoring thread */
#define EXIT_THREAD_STACKSIZE 512       /* Exit thread's stack size */
static TID exit_thread_id;
static BYTE exit_thread_stack[EXIT_THREAD_STACKSIZE];

#endif /* USE_THREAD */

/* function prototypes */
int pophello(int, char *[]);
static void do_popdown(void);
static void do_popup(void);
static THREAD exit_thread(void);
static THREAD mon_thread(void);
static void fatal(const char *, int);
static void loading(void);
static void unloading(void);

/* pophello starts executing here
*/
int main(int argc, char *argv[])
{
    int rval;

#ifdef USE_THREAD

    exit_thread_id = 0;

    /* Get the descriptor of the global infoseg and convert to a far pointer
        so we can get the ID of the current foreground screen group */

    DosGetInfoSeg(&selGlobalSeg, &selLocalSeg);
    pgisInfo = MAKEPGINFOSEG(selGlobalSeg);
    screen_group = pgisInfo->sgCurrent;

    /* We create a unique semaphore for each screen group of the form
        "\sem\popNNNNN", NNNNN = screen group */

    /* first, we build the name of the semaphore */
    strcpy(ssemLoaded_name, ssem_name);
    itoa(screen_group, ssemLoaded_name + strlen(ssemLoaded_name), 10);

    /* Now create the semaphore. We use it to make sure that another
       copy of pophello isn't already running in this screen group. If
       the semaphore does exist, then that's the clue that we are to
       terminate the copy of pophello that is now running. */
    if (rval = DosCreateSem(CSEM_PUBLIC, &ssemLoaded, ssemLoaded_name))
    {
        /* This semaphore name already exists */
        if (rval = DosOpenSem(&ssemLoaded, ssemLoaded_name))
        {
            fatal("System semaphore error #%d", rval);
        }
        /* Signal the previous instance of this program to remove itself */
        DosSemClear(ssemLoaded);
        DosCloseSem(ssemLoaded);
        /* Now remove ourself */
        exit(0);
    }
    else
    {
        /* System semaphore was created successfully */
        /* Indicate that we're memory resident with a semaphore */
        DosSemSet(ssemLoaded);

        /* Check for the existence of the exit thread */
        /* Spin off a thread to look for requests to unload ourself */
        if (rval = DosCreateThread(exit_thread, &exit_thread_id,
            exit_thread_stack + EXIT_THREAD_STACKSIZE))
        {
            fatal("Cannot create exit thread, error #%d", rval);
        }
    }

    /* Show the user that we are loaded */
    loading();

    /* Set up an OS/2 monitor process for console input */

    /* Set up monitor thread id */
    mon_thread_id = 0;

    /* Allocate space for monitor read/write buffers IN THE SAME SEGMENT */
    InBuff = (MONBUFFER *) malloc(MINBUFFSIZE * 2);
    OutBuff = (MONBUFFER *) ((char *)InBuff + MINBUFFSIZE);

    /* Prepare buffer headers for device monitor registration process */
    InBuff->bufflen = MINBUFFSIZE;
    OutBuff->bufflen = MINBUFFSIZE;

    /* Obtain a handle for registering buffers */
    if ((rval = DosMonOpen("KBD$", &hmonHandle))) goto mon_error;

    /* Register the buffers to be used for monitoring */
    if ((rval = DosMonReg(hmonHandle, (PBYTE)InBuff, (PBYTE)OutBuff, FRONT,
        pgisInfo->sgCurrent))) goto mon_error;

#ifdef KBDTRACE
    /* Just call the monitor function to dump keystrokes for debugging */
    mon_thread();
    unloading();
    exit(0);
#else
    /* Start (or resume) a THREAD to monitor the keyboard for the hotkey */
    DosSemClear(&hsemPopup);

    if (rval = DosCreateThread(mon_thread, &mon_thread_id,
        mon_thread_stack + MON_THREAD_STACKSIZE))
    {
        fatal("Cannot create keyboard monitor thread, error #%d\n", rval);
    }
#endif
#endif /* USE_THREAD */

    /* Keyboard monitoring loop - "pop-up" pophello() if hot key detected */
    unload_flag = FALSE;
    for (;;)
    {
#ifdef USE_THREAD
     /* Set the pop-up semaphore, wait until it's cleared by mon_thread() */
        DosSemSetWait(&hsemPopup, SEM_INDEFINITE_WAIT);
#endif
        {
            if (unload_flag)
            { /* exit_thread() is signalling us! */
                unloading();
                exit(0); /* Remove this detached process */
            }
            /* POP-UP !!! */
#ifdef USE_THREAD
            in_main_thread = TRUE;
#endif
            /* Call the main pop-up application */
            pophello(argc, argv);

#ifdef USE_THREAD
            in_main_thread = FALSE;
            nested_hotkey = FALSE;
            /* Unload, if requested, by waking up exit_thread() */
            if (unload_flag) DosSemClear(ssemLoaded);
#else
            unload_flag = TRUE;
#endif
        }
    }

#ifdef USE_THREAD
    mon_error:  /* Keyboard monitor error! */
    DosMonClose(hmonHandle);
    fatal("Error setting up OS/2 monitor: %d\n", rval);
    rval = 1;
#endif

    return rval;
}

#ifdef USE_THREAD

/* Keyboard monitor thread. Monitors the keyboard at high priority,
   Uses the hsemPopup RAM semaphore to control the pop-up mechanism.
*/
static THREAD mon_thread(void)
{
    register int rval;
    KEYPACKET keybuff;      /* One monitor device record for KBD$ */

#ifndef KBDTRACE
    /* Bump our priority up to Time Critical for keyboard monitoring */
    DosSetPrty(PRTYS_THREAD, PRTYC_TIMECRITICAL, 0, mon_thread_id);
#endif
    for (;;)
    {
        keycount = sizeof(keybuff);

        /* Read a keyboard event */
        if (!(rval = DosMonRead((PBYTE)InBuff, DCWW_WAIT, (PBYTE)&keybuff,
            (PUSHORT)&keycount)))
        {
#ifdef KBDTRACE /* Cannot run outside primary thread */
            /* Debug keystroke trace to stderr */
            fprintf(stderr, 
            "mnflags %04x ddflags %04x chChar %02x chScan %02x",
                keybuff.mnflags, keybuff.ddflags,
                keybuff.cp.chChar, keybuff.cp.chScan)
            fprintf (stderr,
            "fbStatus %02x fsState %04x time %08lx\n",
                keybuff.cp.fbStatus, keybuff.cp.fsState,
                keybuff.cp.time);
#endif
            /* Check for hotkey: CTRL-ALT-T */
#define HOTSTATE (CONTROL|ALT)
            if ((keybuff.cp.fsState & HOTSTATE) == HOTSTATE &&
                keybuff.cp.chScan == 0x14)
            {
                if (nested_hotkey) /* ignore it */ ;
                /* Detect a hot key entry while we are popped up */
                else if (in_main_thread) nested_hotkey = TRUE;
                else
                {
#ifdef KBDTRACE
                    return; /* Exit keyboard trace when hot key detected */
#else
                    /* Send semaphore signal to pop up */
                    DosSemClear(&hsemPopup);
#endif
                }
            }
            else DosMonWrite((PBYTE)OutBuff, (PBYTE)&keybuff, keycount);
        }
        else fatal("DosMonRead failure #%d", rval);
    }
}

/* OS/2 thread to exit the program if the ssemLoaded system semaphore
   is cleared by another invocation of POPHELLO.
*/
static THREAD exit_thread(void)
{
    /* Block this thread until we get the semaphore */
    DosSemWait(ssemLoaded, SEM_INDEFINITE_WAIT);
    DosCloseSem(ssemLoaded);
    unload_flag = TRUE;
    DosSemClear(&hsemPopup); /* Wake up mainline code to cause unloading */
    DosSleep(10000L); /* Die sleeping */
}

#endif /* USE_THREAD */

/* pophello main function - entered when the hot-key is depressed
*/
int pophello(int argc, char *argv[])
{
    do_popup();
    printf("\"Hello world,\" says screen group %d", screen_group);
    DosSleep(2000L);
    do_popdown();
    return 0;
}

/* Signal to the user that we are unloading ourself from RAM
*/
static void unloading(void)
{
    do_popup();
    printf("Unloading ...");
    DosSleep(1500L);
    do_popdown();
}

/* Signal to the user that we are loading ourself into RAM
*/
static void loading(void)
{
    do_popup();
    printf("Loading POPHELLO ...\n\n");
    printf("Run POPHELLO to remove from memory.\n");
    DosSleep(2000L);
    do_popdown();
}

/* A detached process cannot access the physical screen normally.
    We must use VioPopUp to gain access to the screen.
*/
static void do_popup(void)
{
    /* Cannot access physical screen directly, use VioPopUp */
    if (!popped_up)
    {
        waitflag = (VP_OPAQUE | VP_WAIT);
        VioPopUp(&waitflag, 0);
        popped_up = TRUE;
    }
}

/* Release our hold on the screen - restoring the previous display
*/
static void do_popdown(void)
{
    if (popped_up)
    {
        VioEndPopUp(0);
    }
    popped_up = FALSE;
}

/* Fatal error handler - Pops up showing the error number
*/
static void fatal(const char *fmt, int errval)
{
    char msg[81];

    do_popup();
    sprintf(msg, fmt, errval);
    printf("### FATAL ERROR IN POPHELLO:\n\n %s\n", msg);
    DosSleep(5000L);
    do_popdown();
    exit(2);
}